/* This library provides basic methods to manipulate pictures, either
 * on screen (FrameBuffers)) or in memmory (OffsreenBuffers). It
 * provides initialization and mmaping of the Linux FrameBuffer,
 * methods to copy pictures between buffers and draw pictures with
 * transparency. */

#ifndef __FBLIB_H__
#define __FBLIB_H__

#include <fcntl.h>
#include <linux/fb.h>
#include <ws/misc.h>

/* Font is a pointer to a font-structure, as created by the bdf2h
 * utility. */
#ifndef DEFINED_STRUCT_FONTINFO_TAG
#define DEFINED_STRUCT_FONTINFO_TAG 1
struct fontinfo_tag
 {
  signed char *data;
  int table[256 - 31];
  int yofs,fh,oy;
 };
#endif
typedef struct fontinfo_tag *Font; 

/* Colours as fed to the drawing functions */
typedef UnsignedChar ColourValue;
typedef struct
 {
  ColourValue blue;
  ColourValue green;
  ColourValue red;
  ColourValue transp;
 } Colour;

/* Coordinates and sizes of graphical objects */
typedef Int Coordinate;
typedef UnsignedInt Size;

/* A FrameBuffer describes a drawing area, either the physical screen,
 * or a virtual area in memmory. It provides the methods to manipulate
 * that area. Do not atempt to use the methods one FrameBuffer on
 * another one. The methods taking several FrameBuffer arguments,
 * should be taken from the FrameBuffer fed as their first
 * argument. */
struct FrameBufferT;
struct FrameBufferT
 {
  char *fbp;
  struct fb_var_screeninfo vinfo;
  struct fb_fix_screeninfo finfo;
  long int screensize;

  /* Returns a pointer to the pixel at location [x, y]. This can be
   * fed to putPixelLocation, and you can add any x incrementation to
   * it (but it's your problem if you run out of the right side of the
   * screen). */
  char *(*pixelLocation)(struct FrameBufferT *fb, Coordinate x, Coordinate y);
  /* Paints the pixel at a specific location (obtained from
   * pixelLocation above) with a specific colour */
  void (*putPixelLocation)(char *location, Colour colour);
  /* Paints one pixel with a colour */
  void (*putPixel)(struct FrameBufferT *fb, Coordinate x, Coordinate y, Colour colour);
  /* Draw a filled rectangle, preserving the old picture according to
   * the transparency specified in the colour */
  void (*drawFilledRectangle)(struct FrameBufferT *fb, Coordinate x, Coordinate y, Size dx, Size dy, Colour colour);
  /* Put a filled rectangle, setting transparency to the one specified
   * in the colour */
  void (*putFilledRectangle)(struct FrameBufferT *fb, Coordinate x, Coordinate y, Size dx, Size dy, Colour colour);
  /* Copy an area from one FrameBuffer to another, or from
   * _non_overlapping_ areas in the same FrameBuffer. */
  void (*copyArea)(struct FrameBufferT *tfb, struct FrameBufferT *ffb, Coordinate tx, Coordinate ty, Coordinate fx, Coordinate fy, Size w, Size h);
  /* Copy an area from one FrameBuffer to another, or between areas in
   * the same FrameBuffer. The areas may overlap. */
  void (*moveArea)(struct FrameBufferT *tfb, struct FrameBufferT *ffb, Coordinate tx, Coordinate ty, Coordinate fx, Coordinate fy, Size w, Size h);
  /* Draws the image of an area of one FrameBuffer onto another,
   * preserving the old image in the destination buffer according to
   * the transparency values of the colours of the pixels in the
   * source buffer. */
  void (*drawOverlay)(struct FrameBufferT *fb, struct FrameBufferT *overlay, Coordinate x, Coordinate y, Coordinate ox, Coordinate oy, Size w, Size h);
  /* Copy an area of one FrameBuffer, drawing the picture of another
   * one in the same way drawOverlay does it, on top of it at the same
   * time, onto a third one. */
  void (*copyAndDrawOverlay)(struct FrameBufferT *tfb, struct FrameBufferT *ffb, struct FrameBufferT *overlay, Coordinate tx, Coordinate ty, Coordinate fx, Coordinate fy, Coordinate ox, Coordinate oy, Size w, Size h);
  size_t (*drawText)(struct FrameBufferT *fb, Coordinate x, Coordinate y, Size w, Size h, Coordinate vx, Coordinate vy, Size vw, Size vh, char *text, Font font, Colour colour);
 };
typedef struct FrameBufferT FrameBuffer;


/* These are the structures describing pixels under various
 * colour-depths. There are mainly two types of pixels - pixels with
 * transparency and pixels with no transparency. */
#define MACcreatPixelFormat3(MACsname, MACcname, MAC0fname, MAC1fname, MAC2fname, MAC0cname, MAC1cname, MAC2cname) \
typedef struct \
 { \
  uint MAC0fname : MACcname##_WIDTH_##MAC0cname; \
  uint MAC1fname : MACcname##_WIDTH_##MAC1cname; \
  uint MAC2fname : MACcname##_WIDTH_##MAC2cname; \
 } __attribute__((packed)) MACsname;

#define MACcreatPixelFormat4(MACsname, MACcname, MAC0fname, MAC1fname, MAC2fname, MAC3fname, MAC0cname, MAC1cname, MAC2cname, MAC3cname) \
typedef struct \
 { \
  uint MAC0fname : MACcname##_WIDTH_##MAC0cname; \
  uint MAC1fname : MACcname##_WIDTH_##MAC1cname; \
  uint MAC2fname : MACcname##_WIDTH_##MAC2cname; \
  uint MAC3fname : MACcname##_WIDTH_##MAC3cname; \
 } __attribute__((packed)) MACsname;

/* Pixel formats. Sizes and offsets (in bits) from LSB of the colour
 * fields. */
#define PIXEL8_WIDTH_BLUE 3
#define PIXEL8_WIDTH_GREEN 2
#define PIXEL8_WIDTH_RED 3
#define PIXEL8_WIDTH_TRANSP 0
#define PIXEL8_OFFSET_BLUE 0
#define PIXEL8_OFFSET_GREEN 3
#define PIXEL8_OFFSET_RED 5
#define PIXEL8_OFFSET_TRANSP 8
MACcreatPixelFormat3(Pixel8, PIXEL8, blue, green, red, BLUE, GREEN, RED)

#define PIXEL16_WIDTH_BLUE 5
#define PIXEL16_WIDTH_GREEN 6
#define PIXEL16_WIDTH_RED 5
#define PIXEL16_WIDTH_TRANSP 0
#define PIXEL16_OFFSET_BLUE 0
#define PIXEL16_OFFSET_GREEN 5
#define PIXEL16_OFFSET_RED 11
#define PIXEL16_OFFSET_TRANSP 16
MACcreatPixelFormat3(Pixel16, PIXEL16, blue, green, red, BLUE, GREEN, RED)

#define PIXEL24_WIDTH_BLUE 8
#define PIXEL24_WIDTH_GREEN 8
#define PIXEL24_WIDTH_RED 8
#define PIXEL24_WIDTH_TRANSP 0
#define PIXEL24_OFFSET_BLUE 0
#define PIXEL24_OFFSET_GREEN 8
#define PIXEL24_OFFSET_RED 16
#define PIXEL24_OFFSET_TRANSP 24
MACcreatPixelFormat3(Pixel24, PIXEL24, blue, green, red, BLUE, GREEN, RED)

#define PIXEL32_WIDTH_BLUE 8
#define PIXEL32_WIDTH_GREEN 8
#define PIXEL32_WIDTH_RED 8
#define PIXEL32_WIDTH_TRANSP 8
#define PIXEL32_OFFSET_BLUE 0
#define PIXEL32_OFFSET_GREEN 8
#define PIXEL32_OFFSET_RED 16
#define PIXEL32_OFFSET_TRANSP 24
MACcreatPixelFormat4(Pixel32, PIXEL32, blue, green, red, transp, BLUE, GREEN, RED, TRANSP)


/* Macros to set the colour-fields of pixels. Used to transfor a
 * Colour struct into a pixel of some format. These are required,
 * since the colour-values in teh colour struct are all 8 bits, but in
 * a pixel the bits allocated to a given colour may be less (for
 * transparency, it may be 0, for the other colours, it may be 5, 6 or
 * 8, in the colourdepths defined here...) */
#define MACsetPixelRedPixel8(MAClocation, MACred)       ((Pixel8 *) (MAClocation))->red    = MACred>>(8 - PIXEL8_WIDTH_RED)
#define MACsetPixelGreenPixel8(MAClocation, MACgreen)   ((Pixel8 *) (MAClocation))->green  = MACgreen>>(8 - PIXEL8_WIDTH_GREEN)
#define MACsetPixelBluePixel8(MAClocation, MACblue)     ((Pixel8 *) (MAClocation))->blue   = MACblue>>(8 - PIXEL8_WIDTH_BLUE)
#define MACsetPixelTranspPixel8(MAClocation, MACtransp)  ({MACtransp;})

#define MACsetPixelRedPixel16(MAClocation, MACred)       ((Pixel16 *) (MAClocation))->red    = MACred>>(8 - PIXEL16_WIDTH_RED)
#define MACsetPixelGreenPixel16(MAClocation, MACgreen)   ((Pixel16 *) (MAClocation))->green  = MACgreen>>(8 - PIXEL16_WIDTH_GREEN)
#define MACsetPixelBluePixel16(MAClocation, MACblue)     ((Pixel16 *) (MAClocation))->blue   = MACblue>>(8 - PIXEL16_WIDTH_BLUE)
#define MACsetPixelTranspPixel16(MAClocation, MACtransp)  ({MACtransp;})

#define MACsetPixelRedPixel24(MAClocation, MACred)       ((Pixel24 *) (MAClocation))->red    = MACred>>(8 - PIXEL24_WIDTH_RED)
#define MACsetPixelGreenPixel24(MAClocation, MACgreen)   ((Pixel24 *) (MAClocation))->green  = MACgreen>>(8 - PIXEL24_WIDTH_GREEN)
#define MACsetPixelBluePixel24(MAClocation, MACblue)     ((Pixel24 *) (MAClocation))->blue   = MACblue>>(8 - PIXEL24_WIDTH_BLUE)
#define MACsetPixelTranspPixel24(MAClocation, MACtransp) ({MACtransp;})

#define MACsetPixelRedPixel32(MAClocation, MACred)       ((Pixel32 *) (MAClocation))->red    = MACred>>(8 - PIXEL32_WIDTH_RED)
#define MACsetPixelGreenPixel32(MAClocation, MACgreen)   ((Pixel32 *) (MAClocation))->green  = MACgreen>>(8 - PIXEL32_WIDTH_GREEN)
#define MACsetPixelBluePixel32(MAClocation, MACblue)     ((Pixel32 *) (MAClocation))->blue   = MACblue>>(8 - PIXEL32_WIDTH_BLUE)
#define MACsetPixelTranspPixel32(MAClocation, MACtransp) ((Pixel32 *) (MAClocation))->transp = MACtransp>>(8 - PIXEL32_WIDTH_TRANSP)

/* Macros to convert from a pixel to a Colour struct */
#define MACgetPixelRedPixel8(MAClocation)                ASL(((ColourValue) ((Pixel8 *) (MAClocation))->red), (8 - PIXEL8_WIDTH_RED))
#define MACgetPixelGreenPixel8(MAClocation)              ASL(((ColourValue) ((Pixel8 *) (MAClocation))->green), (8 - PIXEL8_WIDTH_GREEN))
#define MACgetPixelBluePixel8(MAClocation)               ASL(((ColourValue) ((Pixel8 *) (MAClocation))->blue), (8 - PIXEL8_WIDTH_BLUE))
#define MACgetPixelTranspPixel8(MAClocation)             0

#define MACgetPixelRedPixel16(MAClocation)               ASL(((ColourValue) ((Pixel16 *) (MAClocation))->red), (8 - PIXEL16_WIDTH_RED))
#define MACgetPixelGreenPixel16(MAClocation)             ASL(((ColourValue) ((Pixel16 *) (MAClocation))->green), (8 - PIXEL16_WIDTH_GREEN))
#define MACgetPixelBluePixel16(MAClocation)              ASL(((ColourValue) ((Pixel16 *) (MAClocation))->blue), (8 - PIXEL16_WIDTH_BLUE))
#define MACgetPixelTranspPixel16(MAClocation)            0

#define MACgetPixelRedPixel24(MAClocation)               ASL(((ColourValue) ((Pixel24 *) (MAClocation))->red), (8 - PIXEL24_WIDTH_RED))
#define MACgetPixelGreenPixel24(MAClocation)             ASL(((ColourValue) ((Pixel24 *) (MAClocation))->green), (8 - PIXEL24_WIDTH_GREEN))
#define MACgetPixelBluePixel24(MAClocation)              ASL(((ColourValue) ((Pixel24 *) (MAClocation))->blue), (8 - PIXEL24_WIDTH_BLUE))
#define MACgetPixelTranspPixel24(MAClocation)            0

#define MACgetPixelRedPixel32(MAClocation)               ASL(((ColourValue) ((Pixel32 *) (MAClocation))->red), (8 - PIXEL32_WIDTH_RED))
#define MACgetPixelGreenPixel32(MAClocation)             ASL(((ColourValue) ((Pixel32 *) (MAClocation))->green), (8 - PIXEL32_WIDTH_GREEN))
#define MACgetPixelBluePixel32(MAClocation)              ASL(((ColourValue) ((Pixel32 *) (MAClocation))->blue), (8 - PIXEL32_WIDTH_BLUE))
#define MACgetPixelTranspPixel32(MAClocation)            ASL(((ColourValue) ((Pixel32 *) (MAClocation))->transp), (8 - PIXEL32_WIDTH_TRANSP))


/* Macros used to create multiples of a function, each for a different
 * colour depth. */
#define MACmakeAllFormats(MACname) \
MAC##MACname##Creat(Pixel8) \
MAC##MACname##Creat(Pixel16) \
MAC##MACname##Creat(Pixel24) \
MAC##MACname##Creat(Pixel32)
#define MACmakeAllDepths(MACname) \
MAC##MACname##Creat(8) \
MAC##MACname##Creat(16) \
MAC##MACname##Creat(24) \
MAC##MACname##Creat(32)
#define MACformatFromDepth(MACdepth) Pixel##MACdepth


/* Draws a colour onto a pixel, partly preserving the old colour of
 * the pixel, depending on the transparency of the colour. */
#define MACdrawPixelLocation(MACpixelformat, MAClocation, red, green, blue, transp) \
 do { \
  MACsetPixelRed##MACpixelformat(MAClocation, \
                                 (ColourValue) ((  ((unsigned long int) MACgetPixelRed##MACpixelformat(MAClocation) * (unsigned long int) transp) \
                                                 + (unsigned long int) red * (unsigned long int) (255 - transp)) / 255)); \
  MACsetPixelGreen##MACpixelformat(MAClocation, \
                                   (ColourValue) ((  ((unsigned long int) MACgetPixelGreen##MACpixelformat(MAClocation) * (unsigned long int) transp) \
                                                   + (unsigned long int) green * (unsigned long int) (255 - transp)) / 255)); \
  MACsetPixelBlue##MACpixelformat(MAClocation, \
                                  (ColourValue) ((  ((unsigned long int) MACgetPixelBlue##MACpixelformat(MAClocation) * (unsigned long int) transp) \
                                                  + (unsigned long int) blue * (unsigned long int) (255 - transp)) / 255)); \
 } while(0)


/* Copies the colour of one pixel onto another and draws the colour
 * specified onto it, in the same way as MACdrawPixelLocation. */
#define MACcopyAndDrawPixelLocation(MACpixelformat, MACtlocation, MACflocation, red, green, blue, transp) \
 do { \
  MACsetPixelRed##MACpixelformat(MACtlocation, \
                                 (ColourValue) ((  ((unsigned long int) MACgetPixelRed##MACpixelformat(MACflocation) * (unsigned long int) transp) \
                                                 + (unsigned long int) red * (unsigned long int) (255 - transp)) / 255)); \
  MACsetPixelGreen##MACpixelformat(MACtlocation, \
                                   (ColourValue) ((  ((unsigned long int) MACgetPixelGreen##MACpixelformat(MACflocation) * (unsigned long int) transp) \
                                                   + (unsigned long int) green * (unsigned long int) (255 - transp)) / 255)); \
  MACsetPixelBlue##MACpixelformat(MACtlocation, \
                                  (ColourValue) ((  ((unsigned long int) MACgetPixelBlue##MACpixelformat(MACflocation) * (unsigned long int) transp) \
                                                  + (unsigned long int) blue * (unsigned long int) (255 - transp)) / 255)); \
 } while(0)


/* Calculate a pointer to a pixel within a FrameBuffer. */
#define MACpixelLocation(MACfbp, MACvinfo, MACfinfo, MACx, MACy) ((MACfbp) + ((MACx) + (MACvinfo).xoffset) * ((MACvinfo).bits_per_pixel / 8) + ((MACy) + (MACvinfo).yoffset) * (MACfinfo).line_length)

/* Put a colour onto a pixel of a specific location. */
#define MACputPixelLocation(MACpixelformat, MAClocation, MACred, MACgreen, MACblue, MACtransp) \
 do { \
  MACsetPixelRed##MACpixelformat(MAClocation, MACred); \
  MACsetPixelGreen##MACpixelformat(MAClocation, MACgreen); \
  MACsetPixelBlue##MACpixelformat(MAClocation, MACblue); \
  MACsetPixelTransp##MACpixelformat(MAClocation, MACtransp); \
 } while (0)

/* Put a colour onto a pixel. */
#define MACputPixel(MACpixelformat, MACfbp, MACvinfo, MACfinfo, MACx, MACy, MACred, MACgreen, MACblue, MACtransp) \
  MACputPixelLocation(MACpixelformat, MACpixelLocation(MACfbp, MACvinfo, MACfinfo, MACx, MACy), MACred, MACgreen, MACblue, MACtransp)

/* Reads the colour of a pixel of a specific location. */
#define MACgetPixelLocation(MACpixelformat, MAClocation, MACred, MACgreen, MACblue, MACtransp) \
 do { \
  MACred = MACgetPixelRed##MACpixelformat(MAClocation); \
  MACgreen = MACgetPixelGreen##MACpixelformat(MAClocation); \
  MACblue = MACgetPixelBlue##MACpixelformat(MAClocation); \
  MACtransp = MACgetPixelTransp##MACpixelformat(MAClocation); \
 } while (0)

/* Reads the colour of a pixel. */
#define MACgetPixel(MACpixelformat, MACfbp, MACvinfo, MACfinfo, MACx, MACy, MACred, MACgreen, MACblue, MACtransp) \
  MACgetPixelLocation(MACpixelformat, MACpixelLocation(MACfbp, MACvinfo, MACfinfo, MACx, MACy), MACred, MACgreen, MACblue, MACtransp)

/* Creates a new OffscreenBuffer of a given colour depth. */
#define offscreenBufferNew(MACob, MACpixelformat, MACw, MACh) offscreenBufferNew##MACpixelformat(MACob, MACw, MACh)
#define offscreenBufferNewPixel8(MACob, MACw, MACh) offscreenBufferNewFromDepth(MACob, 8, MACw, MACh)
#define offscreenBufferNewPixel16(MACob, MACw, MACh) offscreenBufferNewFromDepth(MACob, 16, MACw, MACh)
#define offscreenBufferNewPixel24(MACob, MACw, MACh) offscreenBufferNewFromDepth(MACob, 24, MACw, MACh)
#define offscreenBufferNewPixel32(MACob, MACw, MACh) offscreenBufferNewFromDepth(MACob, 32, MACw, MACh)

/* Creates a new OffscreenBuffer of a given colour depth or the depth
 * of an existing Offscreen/FrameBuffer. */
extern int  offscreenBufferNewFromDepth(FrameBuffer *ob, Int depth, Size w, Size h);
extern int  offscreenBufferNewFromExisting(FrameBuffer *ob, FrameBuffer *fb, Size w, Size h);
/* Destroy an OffscreenBuffer, freeing all memory used by it. */
extern void offscreenBufferDestroy(FrameBuffer *ob);
/* Creates or destroys a FrameBuffer. fbfd should be an open
 * filedescriptor to the FrameBuffer device. */
extern int  frameBufferNew(FrameBuffer *fb, int fbfd);
extern void frameBufferDestroy(FrameBuffer *fb);

/* Calculate the width required to draw a text with a given font,
 * without wrapping its lines. That is, the length in pixels of the
 * longest line. */
extern Size drawTextWidth(unsigned char *text, Font font);
/* Calculate the height required to draw a text with a given font and
 * a given width (with line-wrap if required to fit the width). */
extern Size drawTextHeight(Size w, unsigned char *text, Font font);

extern int  frameBufferInit(FrameBuffer *fb, Boolean active);
extern void offscreenBufferSetup(FrameBuffer *ob, Size w, Size h, char *name);
extern int  offscreenBufferSetupFull(FrameBuffer *ob, Int depth, Size w, Size h, char *name);
extern int  offscreenBufferMalloc(FrameBuffer *ob);


#endif /* __FBLIB_H__ */
