/* This library defines widgets, self-positioning/resizing graphical
 * objects. The main ideas of these are that you should never have to
 * explicitly position an object (text, picture, colour-field, etc),
 * nor should you be concerned that some object may not fit in the
 * space avialable to it - all objects can be scrolled. */

#ifndef __WSLIB_H__
#define __WSLIB_H__

#include <sys/types.h>
#include <limits.h>
#include <ws/rhstack.h>
#include <ws/fblib.h>
#include <ws/mdispatchlib.h>
#include <ws/kbdispatchlib.h>
#include <ws/wslib.config.h>
#include <ws/stacktypes.h>


/*
#define CONFIG_RUNTIME_CASTCHECK
*/


typedef Coordinate WidgetSize;
typedef Coordinate WidgetOffset;
#define MAX_INFINITE INT_MAX
#define MIN_UNEXISTENT 0


enum WidgetNumbers
 {
  /* All widgets */
  WN_Widget,
  WN_Container,
  WN_Bin,
  WN_Frame,
  WN_Box,
  WN_Rect,
  WN_Text,
  WN_Image,
  WN_VirtualFrame,
  WN_Input,
  WN_MouseIn,
  WN_KbIn
 };

typedef enum
 {
  /* Widget types and inheritances. The inheritance-values are used
   * for the type-field of widget structs. See ISOFTYPE below for more
   * info. */
  #define WIDGET(name, parent) WT_##name = 1 << WN_##name, WTI_##name = WT_##name | WTI_##parent
  WTI_None = 0,
  WIDGET(Widget,       None),
  WIDGET(Container,    Widget),
  WIDGET(Bin,          Container),
  WIDGET(Frame,        Bin),
  WIDGET(Box,          Container),
  WIDGET(Rect,         Widget),
  WIDGET(Text,         Widget),
  WIDGET(Image,        Widget),
  WIDGET(VirtualFrame, Bin),
  WIDGET(Input,        VirtualFrame),
  WIDGET(MouseIn,      Input),
  WIDGET(KbIn,         Input)
 } WidgetType;

/* Just used for defining the real CONVERT macros below. These are
 * non-checking versions, which might either be used raw, or wrapped
 * with an ISOFTYPE test...*/
#define __CONVERT__(MACtype, MACobj) (*((MACtype *) &(MACobj)))
#define __PCONVERT__(MACtype, MACpoint) (*((MACtype *) MACpoint))
#define __CONVERTP__(MACtype, MACobj) ((MACtype *) &(MACobj))
#define __PCONVERTP__(MACtype, MACpoint) ((MACtype *) MACpoint)

/* This macro checks if a widget is either of a given type, or
 * inherits of that type (thus being cast:able to that type). */
#define ISOFTYPE(MACtype, MACpoint) (__PCONVERT__(Widget, MACpoint).type & WT_##MACtype)

/* These macros casts between widgets. Never cast a widget struct or
 * pointer "by hand" - allways use these macros. If
 * CONFIG_RUNTIME_CASTCHECK, these use ISOFTYPE from above to, at
 * runtime, validate that a cast is only performed to a type of which
 * the widget inherits */
#ifdef CONFIG_RUNTIME_CASTCHECK
#define CONVERT(MACtype, MACobj) ((ISOFTYPE(MACtype, &MACobj) ? 0 : castError("CONVERT", __FILE__, __LINE__, __CONVERT__(Widget, MACobj).type, WT_##MACtype##)), __CONVERT__(MACtype, MACobj))
#define PCONVERT(MACtype, MACpoint) ((ISOFTYPE(MACtype, MACpoint) ? 0 : castError("PCONVERT", __FILE__, __LINE__, __PCONVERT__(Widget, MACpoint).type, WT_##MACtype##)), __PCONVERT__(MACtype, MACpoint))
#define CONVERTP(MACtype, MACobj) ((ISOFTYPE(MACtype, &MACobj) ? 0 : castError("CONVERTP", __FILE__, __LINE__, __CONVERT__(Widget, MACobj).type, WT_##MACtype##)), __CONVERTP__(MACtype, MACobj))
#define PCONVERTP(MACtype, MACpoint) ((ISOFTYPE(MACtype, MACpoint) ? 0 : castError("PCONVERTP", __FILE__, __LINE__, __PCONVERT__(Widget, MACpoint).type, WT_##MACtype##)), __PCONVERTP__(MACtype, MACpoint))
#else
#define CONVERT __CONVERT__
#define PCONVERT __PCONVERT__
#define CONVERTP __CONVERTP__
#define PCONVERTP __PCONVERTP__
#endif

/* Options for widgets. Options is a bit-field. The meaning of the
 * bits varies between widgets, but bits defined for a given widget,
 * have the same meaning for all widgets inheriting from it. Each
 * widget should have a define named WIDGETNAME##OptionsSize, defined
 * to the index of the first bit not defined for that widget. This
 * define is to be used by any widget inheriting from it, to determine
 * which bits to define for its own. */
typedef Int Options;

/* Widget is the base widget. It has no options, and can not be used
 * directly. It must only be inherited from. It defines members used
 * for the drawing of any widget. */
#define WidgetOptionsSize 0
struct WidgetT;
typedef struct WidgetT *WidgetPointer;
#define WidgetPointerCopy(x) x
#define WidgetPointerDelete(x)
struct WidgetT
 {
  WidgetType        type;
  WidgetPointer     parent;
  Options           options;
  VoidPointerStack  userData;
  /* Requested sizes */
  WidgetSize        minWidth;
  WidgetSize        minHeight;
  WidgetSize        maxWidth;
  WidgetSize        maxHeight;
  /* Allocated sizes */
  WidgetSize        width;
  WidgetSize        height;
  /* Position of window to scroll inside */
  WidgetSize        windowX;
  WidgetSize        windowY;
  /* Size of window to scroll inside */
  WidgetSize        windowWidth;
  WidgetSize        windowHeight;
  /* Position of the currently visible area of the window (relative
   * to the left/top of the window) */
  WidgetSize        visibleX;
  WidgetSize        visibleY;
  /* Size of the currently visible area of the window */
  WidgetSize        visibleWidth;
  WidgetSize        visibleHeight;
  /* Scroll position */
  WidgetSize        scrollX;
  WidgetSize        scrollY;

  /* Old position used when moving widgets */
  /* Position of window to scroll inside */
  WidgetSize        oldWindowX;
  WidgetSize        oldWindowY;
  /* Size of window to scroll inside */
  WidgetSize        oldWindowWidth;
  WidgetSize        oldWindowHeight;
  /* Position of the currently visible area of the window (relative
   * to the left/top of the window) */
  WidgetSize        oldVisibleX;
  WidgetSize        oldVisibleY;
  /* Size of the currently visible area of the window */
  WidgetSize        oldVisibleWidth;
  WidgetSize        oldVisibleHeight;
  /* Scroll position */
  WidgetSize        oldScrollX;
  WidgetSize        oldScrollY;

  /* Let the widget calculate the minimum and maximum space it wants */
  void (*calculateRequestedWidth) (WidgetPointer widget);
  void (*calculateRequestedHeight)(WidgetPointer widget);
  /* Let the widget calculate the size of its graphical elements
   * while it knows its own size */
  void (*calculateWidth)          (WidgetPointer widget);
  void (*calculateHeight)         (WidgetPointer widget);
  /* Show the widget (and all its graphical elements, such as
   * child-widgets). */
  void (*show)                    (FrameBuffer *fb, WidgetPointer widget);
  /* Redraw a widget allready drawn, but which has been moved and/or
   * resized */
  void (*move)                    (FrameBuffer *fb, WidgetPointer widget);
  /* Tell a widget (and all children) they are no longer visible. This
   * is used by e.g. the MouseIn and KbIn widgets to remove their
   * keyboard/mouse regions */
  void (*hide)                    (WidgetPointer widget);
  /* Scroll a widget to position x horizontally, and position y
   * vertically. Negative scroll values are scroll values counted from
   * teh other end of the widget. Minus one, since we want to be able
   * to specify "-0". */
  void (*scroll)                  (WidgetPointer widget, WidgetSize x, WidgetSize y);
  /* When a widget's requested sizes have been changed for some
   * reason, this adjusts the requested size of the parent, and its
   * parent, and so on, recursively. It returns the fiorst parent it
   * encounters, which does not change its size in parent, and the
   * number of parents traversed to come to that parent, as return
   * value. This is the upside-down of calculateRequestedWidth. */
  Int (*recalculateParentWidth) (WidgetPointer *parent, WidgetPointer widget);
  Int (*recalculateParentHeight) (WidgetPointer *parent, WidgetPointer widget);
 };
typedef struct WidgetT Widget;
extern void widgetCreat(WidgetPointer this, Options options);
STACKDECL(WidgetPointer)

/* Container is another widget not to be used directly. It should be
 * inherited by all widgets containing child widgets. */
#define ContainerOptionsSize (WidgetOptionsSize + 2)
enum
 {
  /* If CHILDSCROLL_* is set, the child is given as much space as can
   * be given to it, and is forecd to scroll itself if it can not fit
   * in that space. If CHILDSCROLL_* is not set, however, the child is
   * given as much space as it asks for, and the current widget
   * scrolls, if not given enought space. */
  CHILDSCROLL_HORIZONTAL = 1 << WidgetOptionsSize,
  CHILDSCROLL_VERTICAL = 1 << (WidgetOptionsSize + 1)
 } ContainerOptions;
struct ContainerT;
typedef struct ContainerT *ContainerPointer;
struct ContainerT
 {
  Widget  parent;
  Colour  bg;
  /* Set this if you have made changes to the number of children or
   * whatso ever to force a recalculation of X positions and
   * visibilities */
  Boolean childChanges;
  /* Let the container calculate where to show its graphical
   * elements, now when it knows its own size and position */
  void (*calculateXPositions)   (ContainerPointer widget);
  void (*calculateYPositions)   (ContainerPointer widget);
  /* Let the container calculate which of its graphical elements, and
   * how much of them, to show, now when it knows what part of it is
   * visible and the positions of all its elements */
  void (*calculateVisibleWidth) (ContainerPointer widget);
  void (*calculateVisibleHeight)(ContainerPointer widget);
  /* These methods are to be called by an inheriting widget, for each
   * of its children, when its calculateVisible* is called. They
   * calculate the visible width/height of the child passed to
   * them. The visible width/height of the current widget must
   * allready be known, as well as the X/Y coordinate and size of the
   * widget and the child. */
  void (*calculateChildVisibleWidth) (ContainerPointer widget, WidgetPointer child);
  void (*calculateChildVisibleHeight)(ContainerPointer widget, WidgetPointer child);
 };
typedef struct ContainerT Container;
extern void containerCreat(ContainerPointer this, Options options, Colour bg);

/* Bin should not be used directly. It is to be inherited by all
 * widgets with only one child widget. */
#define BinOptionsSize ContainerOptionsSize
struct BinT;
typedef struct BinT *BinPointer;
struct BinT
 {
  Container          parent;
  WidgetPointer      child;
 };
typedef struct BinT Bin;
extern void binCreat(BinPointer this, Options options, WidgetPointer child, Colour bc);

/* Frame puts a single-coloured frame around a single child. The width
 * of the frames of the four sides may be set
 * independently. Optionally, the widget may take up as much space as
 * given to it, thickening the right/bottom borders to fill that
 * space, or all borders (centering the child between the borders). */
#define FrameOptionsSize (ContainerOptionsSize + 4)
enum
 {
  /* let the widget take up as much space given to it, padding with
   * extra border */
  FILL_HORIZONTAL   = 1 << ContainerOptionsSize,
  FILL_VERTICAL     = 1 << (ContainerOptionsSize + 1),
  /* Pad both sides of teh child, not just to its right/bottom, when
   * given too much space */
  CENTER_HORIZONTAL = 1 << (ContainerOptionsSize + 2),
  CENTER_VERTICAL   = 1 << (ContainerOptionsSize + 3)
 } FrameOptions;
struct FrameT;
typedef struct FrameT *FramePointer;
struct FrameT
 {
  Bin                parent;
  WidgetSize         leftMargin;
  WidgetSize         rightMargin;
  WidgetSize         topMargin;
  WidgetSize         bottomMargin;
 };
typedef struct FrameT Frame;
extern void frameCreat(FramePointer this, Options options, WidgetPointer child, WidgetSize leftMargin, WidgetSize rightMargin, WidgetSize topMargin, WidgetSize bottomMargin, Colour bc);

/* A box is a vertical or horizontal list of widgets (padded to their
 * right/bottom to make the width/height of them the same). */
#define BoxOptionsSize (ContainerOptionsSize + 1)
enum
 {
  HORIZONTAL = 1 << ContainerOptionsSize
 } BoxOptions;
struct BoxT;
typedef struct BoxT *BoxPointer;
struct BoxT
 {
  Container          parent;
  Boolean            horizontal;
  WidgetPointerStack children;
  void (*pushChild)(BoxPointer widget, WidgetPointer child);
  void (*popChild)(BoxPointer widget);
 };
typedef struct BoxT Box;
extern void boxCreat(BoxPointer this, Options options, Colour bg);

/* This widget is just a recangle filled with a single colour */
#define RectOptionsSize WidgetOptionsSize
struct RectT;
typedef struct RectT *RectPointer;
struct RectT
 {
  Widget             parent;
  Colour             colour;
 };
typedef struct RectT Rect;
extern void rectCreat(RectPointer this, Options options, Colour colour, WidgetSize minWidth, WidgetSize maxWidth, WidgetSize minHeight, WidgetSize maxHeight);

/* This widgets holds a text. The text may span multiple lines. Font
 * has to be a font generated by bdf2h. */
#define TextOptionsSize (WidgetOptionsSize + 1)
enum
 {
  /* If this option is set, the text will adhere to its min/max width,
   * and wrap within the width allocated to it by its parent. The
   * height will depend on the width available. If unset, the min and
   * max widths will be the length of the longest line, and the height
   * the number of lines. */
  LINE_WRAP = 1 << WidgetOptionsSize,
 } TextOptions;
struct TextT;
typedef struct TextT *TextPointer;
struct TextT
 {
  Widget             parent;
  String             text;
  Font               font;
  Colour             fg;
  Colour             bg;
  /* Requested sizes */
  WidgetSize         minWidth;
  WidgetSize         maxWidth;
  void               (*setText)(TextPointer widget, String text);
#ifdef TEXT_BEAUTIFULLSCROLL
#ifdef TEXT_BACKINGSTORE
  FrameBuffer        backingStore;
#else
  Boolean            textChanged;
#endif
#endif
 };
typedef struct TextT Text;
extern void textCreat(TextPointer this, Options options, String text, Font font, Colour fg, Colour bg, WidgetSize minWidth, WidgetSize maxWidth);

/* This lets you embed an OffscreenBuffer image as a widget (See
 * fblib.h and readlib.h for info on OffscreenBuffers and reading in
 * external pictures). */
#define ImageOptionsSize WidgetOptionsSize
struct ImageT;
typedef struct ImageT *ImagePointer;
struct ImageT
 {
  Widget             parent;
  FrameBuffer       *image;
 };
typedef struct ImageT Image;
extern void imageCreat(ImagePointer this, Options options, FrameBuffer *image);

/* VirtualFrame is a widget to inherit from for widgets with one
 * child, without any other graphical elements (borders and the like),
 * where the child takes up all of the available space. It's main
 * intent is as a baseclass for the Input widget. */
#define VirtualFrameOptionsSize BinOptionsSize
struct VirtualFrameT;
typedef struct VirtualFrameT *VirtualFramePointer;
struct VirtualFrameT
 {
  Bin                parent;
 };
typedef struct VirtualFrameT VirtualFrame;
extern void virtualFrameCreat(VirtualFramePointer this, Options options, WidgetPointer child);

/* Input is a widget to inherit from for all widgets for user input,
 * that is, VirtualFrames with the purpose of grabbing user input
 * (mouse, keyboard, etc...) for the area their child covers. */
#define InputOptionsSize BinOptionsSize
struct InputT;
typedef struct InputT *InputPointer;
struct InputT
 {
  VirtualFrame       parent;
 };
typedef struct InputT Input;
extern void inputCreat(InputPointer this, Options options, WidgetPointer child);

/* A mouseIn is a container that detects mouse-clicks/movements inside
 * its area. The child will allways have the same size as the
 * MouseIn. Only events with the masked buttons pressed/not pressed are
 * collected. Please reffer to mdispatchlib.h and mouselib.h for more
 * information. */
#define MouseInOptionsSize InputOptionsSize
struct MouseInT;
typedef struct MouseInT *MouseInPointer;
typedef void (*MouseInHandler)(MouseInPointer mouseIn, Coordinate x, Coordinate y, MouseButtons buttons, void* data);
typedef void (*MouseInExitHandler)(MouseInPointer mouseIn, void* data);
struct MouseInT
 {
  Input              parent;
  MouseButtons       buttons;
  MouseButtons       mask;
  MDispatcher       *dispatcher;
  MDispatcherEntry  *dispatchEntry;
  MouseInHandler     handler;
  MouseInExitHandler exitHandler;
  MouseInExitHandler changeHandler;
  void              *handlerData;
 };
typedef struct MouseInT MouseIn;
extern void mouseInCreat(MouseInPointer this, Options options, MDispatcher *dispatcher, MouseInHandler handler, MouseInExitHandler exitHandler, MouseInExitHandler changeHandler, MouseButtons buttons, MouseButtons mask, void *handlerData, WidgetPointer child);

/* KbIn is a container that detects keyboard events when having
 * "focus". The child will allways have the same size as the
 * KbIn. Please reffer to kbdispatchlib.h and kblib.h for more
 * information. */
#define KbInOptionsSize InputOptionsSize
struct KbInT;
typedef struct KbInT *KbInPointer;
struct KbInT
 {
  Input                parent;
  KBDispatcher        *dispatcher;
  long                 dispatchEntry;
  void               (*handler)(struct KbInT *kbIn, Key key, void* data);
  void               (*activate)(struct KbInT *kbIn, void* data);
  void               (*deactivate)(struct KbInT *kbIn, void* data);
  void                *handlerData;
 };
typedef struct KbInT KbIn;
typedef void (*KbInHandler)(KbInPointer kbIn, Key key, void* data);
typedef void (*KbInActivateHandler)(KbInPointer kbIn, void* data);
extern void kbInCreat(KbInPointer this, Options options, KBDispatcher *dispatcher, KbInHandler handler, KbInActivateHandler activate, KbInActivateHandler deactivate, void *handlerData, WidgetPointer child);

extern void castError(char *conversion, char *filename, int linenr, WidgetType fromType, WidgetType toType);
extern void topplevelCalculateXPosition    (FrameBuffer *fb, WidgetPointer widget);
extern void topplevelCalculateYPosition    (FrameBuffer *fb, WidgetPointer widget);
extern void topplevelCalculateWidth        (FrameBuffer *fb, WidgetPointer widget);
extern void topplevelCalculateHeight       (FrameBuffer *fb, WidgetPointer widget);
extern void topplevelCalculateVisibleWidth (FrameBuffer *fb, WidgetPointer widget);
extern void topplevelCalculateVisibleHeight(FrameBuffer *fb, WidgetPointer widget);
extern void topplevelShow                  (FrameBuffer *fb, WidgetPointer widget);

/* These are convienence macros, just to automate some common tasks */

#define widgetScrolledV(MACwidget, MAClevel) \
 do { \
  if (vp.level < MAClevel) \
   { vp.widget = MACwidget; vp.level = MAClevel; } \
 } while (0)

#define widgetScrolledH(MACwidget, MAClevel) \
 do { \
  if (hp.level < MAClevel) \
   { hp.widget = MACwidget; hp.level = MAClevel; } \
 } while (0)

#define widgetUpdated(MACwidget, MAClevel) \
 do { \
  Int depth; \
  WidgetPointer parent; \
  \
  if (hp.level < MAClevel) \
   { \
    depth = PCONVERT(Widget, MACwidget).recalculateParentWidth(&parent, PCONVERTP(Widget, MACwidget)) + MAClevel; \
    PCONVERT(Widget, parent).calculateWidth(PCONVERTP(Widget, parent)); \
    if (hp.level < depth) \
     { hp.widget = parent; hp.level = depth; } \
   } \
  if (vp.level < MAClevel) \
   { \
    depth = PCONVERT(Widget, MACwidget).recalculateParentHeight(&parent, PCONVERTP(Widget, MACwidget)) + MAClevel; \
    PCONVERT(Widget, parent).calculateHeight(PCONVERTP(Widget, parent)); \
    if (vp.level < depth) \
     { vp.widget = parent; vp.level = depth; } \
   } \
 } while (0)

#define widgetUpdate(MACwidget, MAClevel) \
 do { \
  if (PCONVERT(Widget, MACwidget).parent) \
   { \
    Int depth; \
    WidgetPointer parent; \
    \
    if (hp.level < MAClevel) \
     { \
      PCONVERT(Widget, MACwidget).calculateRequestedWidth(PCONVERTP(Widget, MACwidget)); \
      depth = PCONVERT(Widget, PCONVERT(Widget, MACwidget).parent).recalculateParentWidth(&parent, PCONVERTP(Widget, PCONVERT(Widget, MACwidget).parent)) + 1 + MAClevel; \
      PCONVERT(Widget, parent).calculateWidth(PCONVERTP(Widget, parent)); \
      if (hp.level < depth) \
       { hp.widget = parent; hp.level = depth; } \
     } \
    if (vp.level < MAClevel) \
     { \
      PCONVERT(Widget, MACwidget).calculateRequestedHeight(PCONVERTP(Widget, MACwidget)); \
      depth = PCONVERT(Widget, PCONVERT(Widget, MACwidget).parent).recalculateParentHeight(&parent, PCONVERTP(Widget, PCONVERT(Widget, MACwidget).parent)) + 1 + MAClevel; \
      PCONVERT(Widget, parent).calculateHeight(PCONVERTP(Widget, parent)); \
      if (vp.level < depth) \
       { vp.widget = parent; vp.level = depth; } \
     } \
   } \
  else \
   { \
    if (hp.level < MAClevel) \
     { \
      PCONVERT(Widget, MACwidget).calculateRequestedWidth(PCONVERTP(Widget, MACwidget)); \
      PCONVERT(Widget, MACwidget).calculateWidth(PCONVERTP(Widget, MACwidget)); \
      hp.widget = MACwidget; hp.level = MAClevel; \
     } \
    if (vp.level < MAClevel) \
     { \
      PCONVERT(Widget, MACwidget).calculateRequestedHeight(PCONVERTP(Widget, MACwidget)); \
      PCONVERT(Widget, MACwidget).calculateHeight(PCONVERTP(Widget, MACwidget)); \
      vp.widget = MACwidget; vp.level = MAClevel; \
     } \
   } \
 } while (0)

typedef struct
 {
  Int level;
  WidgetPointer widget;
 } UpdateParent;

#define widgetUpdateClosure(MACupdates) \
 do { \
  UpdateParent vp = {-1, NULL}; \
  UpdateParent hp = {-1, NULL}; \
  \
  MACupdates \
  if (vp.level < hp.level) \
   vp.widget = hp.widget; \
  if (vp.widget) \
   { \
    if (ISOFTYPE(Container, vp.widget)) \
     { \
      PCONVERT(Container, vp.widget).calculateXPositions(PCONVERTP(Container, vp.widget)); \
      PCONVERT(Container, vp.widget).calculateVisibleWidth(PCONVERTP(Container, vp.widget)); \
      PCONVERT(Container, vp.widget).calculateYPositions(PCONVERTP(Container, vp.widget)); \
      PCONVERT(Container, vp.widget).calculateVisibleHeight(PCONVERTP(Container, vp.widget)); \
     } \
    PCONVERT(Widget, vp.widget).move(&fb, PCONVERTP(Widget, vp.widget)); \
   } \
 } while (0)

#endif /* __WSLIB_H__ */
